
use tera::Tera;
use serde::Deserialize;

use std::fmt;

use crate::mycelium::MycProtoEngine;
use crate::mycelium::read_toml_from_file;
use crate::mycelium::TomlError;

#[derive(Deserialize, Clone)]
pub struct TemplateEngineLoader {
	pub template_location: String,
	pub extra_config_location: Option<String>,
}

impl TemplateEngineLoader {

	pub fn new(
		template_location: String,
		extra_config_location: Option<String>
	) -> Self {
		Self {
			template_location: template_location,
			extra_config_location: extra_config_location,
		}
	}

	pub fn cli_template_location(mut self, location: Option<String>) -> Self {
		if let Some(location) = location {
			self.template_location = location;
		}
		self
	}

	pub fn cli_extra_config_location(mut self, location: Option<String>) -> Self {
		if let Some(location) = location {
			self.extra_config_location = Some(location);
		}
		self
	}

	// Returns the template base directory.
	//
	// Contructed by ensuring the template_location ends in a "/".
	pub fn base_dir(&self) -> String {
		if self.template_location.ends_with("/") {
			self.template_location.clone()
		} else {
			self.template_location.clone()+"/"
		}
	}

	pub fn load_templates(
		&self
	) -> Result<MycProtoEngine,TemplateEngineLoaderError> {
		let template_base_dir = self.base_dir();
		let template_extra_config_res = match &self.extra_config_location {
			Some(path) => read_toml_from_file(path),
			None => {
				read_toml_from_file(&(template_base_dir.clone()+"extra.toml"))
			}
		};
		let template_extra_config = match template_extra_config_res {
			Ok(c) => Some(c),
			Err(e) => match &e {
				TomlError::FileError{..} => {
					// Only fatal if the file was explicitly requested.
					// An implicit request could also mean that
					// the template doesn't need a config file.
					if self.extra_config_location.is_some() {
						return Err(TemplateEngineLoaderError::TomlError(e));
					}
					None
				},
				TomlError::ParseError{..} => {
					return Err(TemplateEngineLoaderError::TomlError(e));
				}
			},
		};
		let template_glob = template_base_dir.clone()+"*";
		println!("Parsing Templates from '{}' ...", &template_glob);
		let res = Tera::new((template_glob).as_str());
		let tera = match res {
			Ok(t) => t,
			Err(e) => {
				return Err(TemplateEngineLoaderError::TemplateParseError{
					path: template_glob,
					tera_error: e,
				});
			}
		};
		Ok(MycProtoEngine {
			tera: tera,
			template_config: template_extra_config,
		})
	}
}

pub enum TemplateEngineLoaderError {
	TomlError(TomlError),
	TemplateParseError{path: String, tera_error: tera::Error },
}

impl fmt::Display for TemplateEngineLoaderError {
	fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
		match self {
			Self::TomlError(e) =>
				write!(f,"Error with template extra configuration:\n(Depending on you template the file not being found might be okay)\n{e}"),
			Self::TemplateParseError{path, tera_error} =>
				write!(f,"Error parsing template '{path}':\n{tera_error}"),
		}
	}
}
